// Charlie Cole 2015
// HDMI output for Neo Geo MVS
//   Originally based on fpga4fun.com HDMI/DVI sample code (c) fpga4fun.com & KNJN LLC 2013
//   Added Neo Geo MVS input, scan doubling, HDMI data packets and audio
//   Offers fake scanline generation (select via button)
//		0: Line doubled but even lines are half brightness
//		1: Only show even lines (odd lines are black)
//		2: Line doubled

module HDMIDirectV(
	input pixclk,
	input pixclk_en,
	input pixclk72,
	input pixclk72n,

	input hSync,
	input vSync,
	input blank,
	
	input [7:0] red,
	input [7:0] green,
	input [7:0] blue,
	
	
	input audioLR,
	input audioClk,
	input audioData,
	
	input HBlankStart,
	input HBlankEnd,

	//output [2:0] TMDS,
	output TMDS_2,
	output TMDS_1,
	output TMDS_0,
	output TMDS_clock
	
	//output [2:0] TMDSp, TMDSn,
	//output TMDSp_clock, TMDSn_clock

);

////////////////////////////////////////////////////////////////////////
// User configuration defines

//`define YM3016		// Comment out for BU9480F (chip found on newer boards)
//`define SPLASH_SCREEN
//`define BAD_SYNC_DETECT
`define DELAY_UNTIL_SPLASH_SCREEN	240	// In frames (/60 for seconds)

////////////////////////////////////////////////////////////////////////

// Defines to do with video signal generation
`define DISPLAY_WIDTH			720
`define DISPLAY_HEIGHT			480
`define FULL_WIDTH				858
`define FULL_HEIGHT				525
`define H_FRONT_PORCH			16
`define H_SYNC						62 
`define V_FRONT_PORCH			9
`define V_SYNC 					6

`define NEOGEO_DISPLAY_WIDTH	640
`define NEOGEO_DISPLAY_HEIGHT	448
`define NEOGEO_FULL_WIDTH		768
`define NEOGEO_FULL_HEIGHT		528
`define NEOGEO_VSYNC_LENGTH	80

`define CENTERING_X				((`DISPLAY_WIDTH-`NEOGEO_DISPLAY_WIDTH)/2)	// For centering NeoGeo's 4:3 screen
`define CENTERING_Y				((`DISPLAY_HEIGHT-`NEOGEO_DISPLAY_HEIGHT)/2)	// Should be multiple of 8

// Defines to do with data packet sending
`define DATA_START		4//(`DISPLAY_WIDTH+4) // Need 4 cycles of control data first
`define DATA_PREAMBLE	8
`define DATA_GUARDBAND	2 
`define DATA_SIZE			32
`define VIDEO_PREAMBLE	8
`define VIDEO_GUARDBAND	2
`define CTL_END			0 //(`FULL_WIDTH-`VIDEO_PREAMBLE-`VIDEO_GUARDBAND)

wire clk_TMDS = pixclk72;
wire clk_TMDSn = pixclk72n;

////////////////////////////////////////////////////////////////////////
// CounterX, CounterY reconstruction
////////////////////////////////////////////////////////////////////////
reg [9:0] CounterX, CounterY;
reg [9:0] CounterHBlankStart, CounterHBlankEnd;
reg hsync_t1, vsync_t1;

initial
begin
	CounterX=0;
	CounterY=0;
end

always @(posedge pixclk)

begin
	if(pixclk_en) begin
		hsync_t1 <= hSync;
		vsync_t1 <= vSync;
		//hSync <= !((CounterX>=(`DISPLAY_WIDTH+`H_FRONT_PORCH)) && (CounterX<(`DISPLAY_WIDTH+`H_FRONT_PORCH+`H_SYNC)));
		if(hsync_t1 == 1 && hSync == 0) begin //falling hSync
			CounterX <= 10'd738;
		end
		else if(CounterX==(`FULL_WIDTH-1)) begin
			if (CounterY==(`FULL_HEIGHT-1)) begin
				CounterY <= 0;
				CounterX <= 0;
			end
			else begin
				CounterY <= CounterY+1;
				CounterX <= 0;
			end
		end
		else begin
			CounterX <= CounterX + 1;
		end
	end
	
end

always @(posedge pixclk)

begin
	if(pixclk_en) begin
		//CounterHBlankStart
		if(HBlankStart==1) begin
			CounterHBlankStart <= 0;
		end
		else if (CounterHBlankStart < 200) begin
			CounterHBlankStart <= CounterHBlankStart + 1;
		end
		else begin
			CounterHBlankStart <= CounterHBlankStart;
		end
		
		//CounterHBlankEnd
		if(HBlankEnd==1) begin
			CounterHBlankEnd <= 0;
		end
		else if (CounterHBlankEnd < 200) begin
			CounterHBlankEnd <= CounterHBlankEnd + 1;
		end
		else begin
			CounterHBlankEnd <= CounterHBlankEnd;
		end
	end
end		


////////////////////////////////////////////////////////////////////////
// Neo Geo audio input
////////////////////////////////////////////////////////////////////////

reg [15:0] audioInput [1:0];
reg [15:0] curSampleL;
reg [15:0] curSampleR;

initial
begin
	audioInput[0]=0;
	audioInput[1]=0;
	curSampleL=0;
	curSampleR=0;
end

`ifdef YM3016
always @(negedge audioClk)
begin
	audioInput[0]<=(audioInput[0]>>1)|(audioData<<15);
	audioInput[1]<=(audioInput[1]>>1)|(audioData<<15);
end
always @(negedge audioLR)  begin curSampleL<=audioInput[0]-16'h8000; end
always @(negedge audioLR2) begin curSampleR<=audioInput[1]-16'h8000; end
`else // BU9480F
always @(posedge audioClk) if (audioLR) audioInput[0]<=(audioInput[0]<<1)|audioData; else audioInput[1]<=(audioInput[1]<<1)|audioData;
always @(negedge audioLR) begin curSampleL<=audioInput[0]; curSampleR<=audioInput[1]; end
`endif

////////////////////////////////////////////////////////////////////////
// HDMI audio packet generator
////////////////////////////////////////////////////////////////////////

// Timing for 32KHz audio at 27MHz
//`define AUDIO_TIMER_ADDITION	4
//`define AUDIO_TIMER_LIMIT		3375

// Timing for 48KHz audio at 27MHz
`define AUDIO_TIMER_ADDITION	2
`define AUDIO_TIMER_LIMIT		1124


//localparam [191:0] channelStatus = 192'hc203004004; // 32KHz 16-bit LPCM audio //3 indicates 32KHz
localparam [191:0] channelStatus = 192'hc202004004; // 48KHz 16-bit LPCM audio
reg [23:0] audioPacketHeader;
reg [55:0] audioSubPacket[3:0];
reg [7:0] channelStatusIdx;
reg [11:0] audioTimer;
reg [9:0] audioSamples;
reg [1:0] samplesHead;
reg sendRegenPacket;

initial
begin
	audioPacketHeader=0;
	audioSubPacket[0]=0;
	audioSubPacket[1]=0;
	audioSubPacket[2]=0;
	audioSubPacket[3]=0;
	channelStatusIdx=0;
	audioTimer=0;
	audioSamples=0;
	samplesHead=0;
	sendRegenPacket=0;
end

task AudioPacketGeneration;
	begin
		// Buffer up an audio sample every 750 pixel clocks (32KHz output from 24MHz pixel clock)
		// Don't add to the audio output if we're currently sending that packet though
		if (!(
			//CounterX>=(`DATA_START+`DATA_PREAMBLE+`DATA_GUARDBAND+`DATA_SIZE) &&
			//CounterX<(`DATA_START+`DATA_PREAMBLE+`DATA_GUARDBAND+`DATA_SIZE+`DATA_SIZE)
			CounterHBlankStart>=(`DATA_START+`DATA_PREAMBLE+`DATA_GUARDBAND+`DATA_SIZE) &&
			CounterHBlankStart<(`DATA_START+`DATA_PREAMBLE+`DATA_GUARDBAND+`DATA_SIZE+`DATA_SIZE)
		)) begin
			if (audioTimer>=`AUDIO_TIMER_LIMIT) begin
			
				// curSample
				//curSampleL <= CounterX;
				//curSampleR <= CounterY;
			
				audioTimer<=audioTimer-`AUDIO_TIMER_LIMIT+`AUDIO_TIMER_ADDITION;
				audioPacketHeader<=audioPacketHeader|24'h000002|((channelStatusIdx==0?24'h100100:24'h000100)<<samplesHead);
				audioSubPacket[samplesHead]<=((curSampleL<<8)|(curSampleR<<32)
									|((^curSampleL)?56'h08000000000000:56'h0)		// parity bit for left channel
									|((^curSampleR)?56'h80000000000000:56'h0))	// parity bit for right channel
									^(channelStatus[channelStatusIdx]?56'hCC000000000000:56'h0); // And channel status bit and adjust parity
				if (channelStatusIdx<191)
					channelStatusIdx<=channelStatusIdx+1;
				else
					channelStatusIdx<=0;
				samplesHead<=samplesHead+1;
				audioSamples<=audioSamples+1;
				if (audioSamples[4:0]==0)
					sendRegenPacket<=1;
			end else begin
				audioTimer<=audioTimer+`AUDIO_TIMER_ADDITION;
			end
		end else begin
			audioTimer<=audioTimer+`AUDIO_TIMER_ADDITION;
			samplesHead<=0;
		end
	end
endtask

////////////////////////////////////////////////////////////////////////
// Error correction code generator
// Generates error correction codes needed for verifying HDMI packets
////////////////////////////////////////////////////////////////////////

function [7:0] ECCcode;	// Cycles the error code generator
	input [7:0] code;
	input bita;
	input passthroughData;
	begin
		ECCcode = (code<<1) ^ (((code[7]^bita) && passthroughData)?(1+(1<<6)+(1<<7)):0);
	end
endfunction

task ECCu;
	output outbit;
	inout [7:0] code;
	input bita;
	input passthroughData;
	begin
		outbit <= passthroughData?bita:code[7];
		code <= ECCcode(code, bita, passthroughData);
	end
endtask

task ECC2u;
	output outbita;
	output outbitb;
	inout [7:0] code;
	inout [7:0] code_a; //added, because else recursive function call needed
	input bita;
	input bitb;
	input passthroughData;	
	begin
		outbita <= passthroughData?bita:code[7];
		outbitb <= passthroughData?bitb:(code[6]^(((code[7]^bita) && passthroughData)?1'b1:1'b0));
		code_a  = ECCcode(code, bita, passthroughData); 
		code <= ECCcode(code_a, bitb, passthroughData);
		//code <= ECCcode(ECCcode(code, bita, passthroughData), bitb, passthroughData);
	end
endtask

////////////////////////////////////////////////////////////////////////
// Packet sending
// During hsync periods send audio data and infoframe data packets
////////////////////////////////////////////////////////////////////////

reg [3:0] dataChannel0;
reg [3:0] dataChannel1;
reg [3:0] dataChannel2;
reg [23:0] packetHeader;
reg [55:0] subpacket[3:0];
reg [7:0] bchHdr;
reg [7:0] bchCode [3:0];
reg [7:0] bchCode_a [3:0];
reg [4:0] dataOffset;
reg [3:0] preamble;
reg tercData;
reg dataGuardBand;
reg videoGuardBand;

initial
begin
	dataChannel0=0;
	dataChannel1=0;
	dataChannel2=0;
	packetHeader=0;
	subpacket[0]=0;
	subpacket[1]=0;
	subpacket[2]=0;
	subpacket[3]=0;
	bchHdr=0;
	bchCode[0]=0;
	bchCode[1]=0;
	bchCode[2]=0;
	bchCode[3]=0;
	dataOffset=0;
	preamble=0;
	tercData=0;
	dataGuardBand=0;
	videoGuardBand=0;
end

task SendPacket;
	inout [32:0] pckHeader;
	inout [55:0] pckData0;
	inout [55:0] pckData1;
	inout [55:0] pckData2;
	inout [55:0] pckData3;
	input firstPacket;
begin
	dataChannel0[0]<=hSync;
	dataChannel0[1]<=vSync;
	dataChannel0[3]<=(!firstPacket || dataOffset)?1'b1:1'b0;
	ECCu(dataChannel0[2], bchHdr, pckHeader[0], dataOffset<24?1'b1:1'b0);
	ECC2u(dataChannel1[0], dataChannel2[0], bchCode[0], bchCode_a[0], pckData0[0], pckData0[1], dataOffset<28?1'b1:1'b0);
	ECC2u(dataChannel1[1], dataChannel2[1], bchCode[1], bchCode_a[1], pckData1[0], pckData1[1], dataOffset<28?1'b1:1'b0);
	ECC2u(dataChannel1[2], dataChannel2[2], bchCode[2], bchCode_a[2], pckData2[0], pckData2[1], dataOffset<28?1'b1:1'b0);
	ECC2u(dataChannel1[3], dataChannel2[3], bchCode[3], bchCode_a[3], pckData3[0], pckData3[1], dataOffset<28?1'b1:1'b0);
	pckHeader<=pckHeader[23:1];
	pckData0<=pckData0[55:2];
	pckData1<=pckData1[55:2];
	pckData2<=pckData2[55:2];
	pckData3<=pckData3[55:2];
	dataOffset<=dataOffset+5'b1;
end
endtask

always @(posedge pixclk)
begin
	if(pixclk_en) begin
		AudioPacketGeneration();
		// Start sending audio data if we're in the right part of the hsync period
		if (CounterHBlankStart>=`DATA_START)
		begin
			if (CounterHBlankStart<(`DATA_START+`DATA_PREAMBLE))
			begin
				// Send the data period preamble
				// A nice "feature" of my test monitor (GL2450) is if you comment out
				// this line you see your data next to your image which is useful for
				// debugging
				preamble<='b0101;
			end
			else if (CounterHBlankStart<(`DATA_START+`DATA_PREAMBLE+`DATA_GUARDBAND))
			begin
				// Start sending leading data guard band
				tercData<=1;
				dataGuardBand<=1;
				dataChannel0<={1'b1, 1'b1, vSync, hSync};
				preamble<=0;
				// Set up the first of the packets we'll send
				if (sendRegenPacket) begin
					packetHeader<=24'h000001;	// audio clock regeneration packet
//					subpacket[0]<=56'h00100078690000;	// N=0x1000 CTS=0x6978 (27MHz pixel clock -> 32KHz audio clock)
//					subpacket[1]<=56'h00100078690000;
//					subpacket[2]<=56'h00100078690000;
//					subpacket[3]<=56'h00100078690000;
					subpacket[0]<=56'h00180060690000;	// N=0x1800 CTS=0x6960 (27MHz pixel clock -> 48KHz audio clock)
					subpacket[1]<=56'h00180060690000;
					subpacket[2]<=56'h00180060690000;
					subpacket[3]<=56'h00180060690000;
					if (CounterHBlankStart==(`DATA_START+`DATA_PREAMBLE+`DATA_GUARDBAND-1))
						sendRegenPacket<=0;
				end else begin
					if (!CounterY[0]) begin
						packetHeader<=24'h0D0282;	// infoframe AVI packet
						// Byte0: Checksum
						// Byte1: 10 = 0(Y1:Y0=0 RGB)(A0=1 No active format)(B1:B0=00 No bar info)(S1:S0=00 No scan info)
						// Byte2: 2A = (C1:C0=0 No colorimetry)(M1:M0=2 16:9)(R3:R0=A 16:9)
						// Byte3: 00 = 0(SC1:SC0=0 No scaling)
						// Byte4: 03 = 0(VIC6:VIC0=3 720x480p)
						// Byte5: 00 = 0(PR5:PR0=0 No repeation)
						//subpacket[0]<=56'h000003002A1032;
						subpacket[0]<=56'h000002002A1032; //480p 16:9
						//subpacket[0]<=56'h001506002A1032; //576i pr=2
						subpacket[1]<=56'h00000000000000;
					end else begin
						packetHeader<=24'h0A0184;	// infoframe audio packet
						// Byte0: Checksum
						// Byte1: 11 = (CT3:0=1 PCM)0(CC2:0=1 2ch)
						// Byte2: 00 = 000(SF2:0=0 As stream)(SS1:0=0 As stream)
						// Byte3: 00 = LPCM doesn't use this
						// Byte4-5: 00 Multichannel only (>2ch)
						subpacket[0]<=56'h00000000001160;
						subpacket[1]<=56'h00000000000000;
					end
					subpacket[2]<=56'h00000000000000;
					subpacket[3]<=56'h00000000000000;
				end
			end
			else if (CounterHBlankStart<(`DATA_START+`DATA_PREAMBLE+`DATA_GUARDBAND+`DATA_SIZE))
			begin
				dataGuardBand<=0;
				// Send first data packet (Infoframe or audio clock regen)
				SendPacket(packetHeader, subpacket[0], subpacket[1], subpacket[2], subpacket[3], 1);
			end
			else if (CounterHBlankStart<(`DATA_START+`DATA_PREAMBLE+`DATA_GUARDBAND+`DATA_SIZE+`DATA_SIZE))
			begin
				// Send second data packet (audio data)
				SendPacket(audioPacketHeader, audioSubPacket[0], audioSubPacket[1], audioSubPacket[2], audioSubPacket[3], 0);
			end
			else if (CounterHBlankStart<(`DATA_START+`DATA_PREAMBLE+`DATA_GUARDBAND+`DATA_SIZE+`DATA_SIZE+`DATA_GUARDBAND))
			begin
				// Trailing guardband for data period
				dataGuardBand<=1;
				dataChannel0<={1'b1, 1'b1, vSync, hSync};	
			end
			else
			begin
				// Back to normal DVI style control data
				tercData<=0;
				dataGuardBand<=0;
			end
		end
		// After we've sent data packets we need to do the video preamble and
		// guardband just before sending active video data
//		if (CounterHBlankEnd>=(`CTL_END+`VIDEO_PREAMBLE))
//		begin
//			preamble<=0;
//			videoGuardBand<=1;
//		end
//		else if (CounterHBlankEnd>=(`CTL_END))
//		begin
//			preamble<='b0001;
//		end
//		else
//		begin
//			videoGuardBand<=0;
//		end
		if (CounterHBlankEnd>=(`CTL_END+`VIDEO_PREAMBLE+`VIDEO_GUARDBAND))
		begin
			videoGuardBand<=0;
		end
		else if (CounterHBlankEnd>=(`CTL_END+`VIDEO_PREAMBLE))
		begin
			preamble<=0;
			videoGuardBand<=1;
		end
		else if (CounterHBlankEnd>=(`CTL_END))
		begin
			preamble<='b0001;
		end
	end
end

////////////////////////////////////////////////////////////////////////
// HDMI encoder
// Encodes video data (TMDS) or packet data (TERC4) ready for sending
////////////////////////////////////////////////////////////////////////

reg tercDataDelayed [1:0];
reg videoGuardBandDelayed [1:0];
reg dataGuardBandDelayed [1:0];

initial
begin
	tercDataDelayed[0]=0;
	tercDataDelayed[1]=0;
	videoGuardBandDelayed[0]=0;
	videoGuardBandDelayed[1]=0;
	dataGuardBandDelayed[0]=0;
	dataGuardBandDelayed[1]=0;
end

always @(posedge pixclk)
begin
	if(pixclk_en) begin
		// Cycle 1
		tercDataDelayed[0]<=tercData;	// To account for delay through encoder
		videoGuardBandDelayed[0]<=videoGuardBand;
		dataGuardBandDelayed[0]<=dataGuardBand;
		// Cycle 2
		tercDataDelayed[1]<=tercDataDelayed[0];
		videoGuardBandDelayed[1]<=videoGuardBandDelayed[0];
		dataGuardBandDelayed[1]<=dataGuardBandDelayed[0];
	end
end

wire [9:0] TMDS_red, TMDS_green, TMDS_blue;
TMDS_encoder encode_R(.clk(pixclk), .pixclk_en(pixclk_en), .VD(red  ), .CD(preamble[3:2]), .VDE(!blank), .TMDS(TMDS_red));
TMDS_encoder encode_G(.clk(pixclk), .pixclk_en(pixclk_en), .VD(green), .CD(preamble[1:0]), .VDE(!blank), .TMDS(TMDS_green));
TMDS_encoder encode_B(.clk(pixclk), .pixclk_en(pixclk_en), .VD(blue ), .CD({vSync,hSync}), .VDE(!blank), .TMDS(TMDS_blue));

//TDMS_encoder TDMS_encoder_red(.clk(pixclk), .clk_en(pixclk_en), .data(red),   .c(preamble[3:2]),.blank(blank),.encoded(TMDS_red));
//TDMS_encoder TDMS_encoder_green(.clk(pixclk), .clk_en(pixclk_en), .data(green),   .c(preamble[1:0]),.blank(blank),.encoded(TMDS_green));
//TDMS_encoder TDMS_encoder_blue(.clk(pixclk), .clk_en(pixclk_en), .data(blue),   .c({vSync,hSync}),.blank(blank),.encoded(TMDS_blue));



wire [9:0] TERC4_red, TERC4_green, TERC4_blue;
TERC4_encoder encode_R4(.clk(pixclk), .clk_en(pixclk_en), .data(dataChannel2), .TERC(TERC4_red));
TERC4_encoder encode_G4(.clk(pixclk), .clk_en(pixclk_en), .data(dataChannel1), .TERC(TERC4_green));
TERC4_encoder encode_B4(.clk(pixclk), .clk_en(pixclk_en), .data(dataChannel0), .TERC(TERC4_blue));

////////////////////////////////////////////////////////////////////////
// HDMI data serialiser
// Outputs the encoded video data as serial data across the HDMI bus
////////////////////////////////////////////////////////////////////////

reg [3:0] TMDS_mod10;  // modulus 10 counter
reg [9:0] TMDS_shift_red, TMDS_shift_green, TMDS_shift_blue;
reg [9:0] TMDS_shift_red_delay, TMDS_shift_green_delay, TMDS_shift_blue_delay;
reg [9:0] TMDS_shift_clock;

initial
begin
  TMDS_mod10=0;
  TMDS_shift_red=0;
  TMDS_shift_green=0;
  TMDS_shift_blue=0;
  TMDS_shift_red_delay=0;
  TMDS_shift_green_delay=0;
  TMDS_shift_blue_delay=0;
end

always @(posedge pixclk)
begin
	if(pixclk_en) begin
		//invert the TMDS pairs
		TMDS_shift_red_delay<=(videoGuardBandDelayed[1] ? 10'b1011001100 : (dataGuardBandDelayed[1] ? 10'b0100110011 : (tercDataDelayed[1] ? TERC4_red : TMDS_red)));
		TMDS_shift_green_delay<=~((dataGuardBandDelayed[1] || videoGuardBandDelayed[1]) ? 10'b0100110011 : (tercDataDelayed[1] ? TERC4_green : TMDS_green));
		TMDS_shift_blue_delay<=~(videoGuardBandDelayed[1] ? 10'b1011001100 : (tercDataDelayed[1] ? TERC4_blue : TMDS_blue));
	end
end

always @(posedge clk_TMDS)
begin
	TMDS_shift_red   <= (TMDS_mod10==4'd8) ? TMDS_shift_red_delay   : TMDS_shift_red  [9:2];
	TMDS_shift_green <= (TMDS_mod10==4'd8) ? TMDS_shift_green_delay : TMDS_shift_green[9:2];
	TMDS_shift_blue  <= (TMDS_mod10==4'd8) ? TMDS_shift_blue_delay  : TMDS_shift_blue [9:2];
	TMDS_shift_clock <= (TMDS_mod10==4'd8) ? 10'b0000011111  : TMDS_shift_clock [9:2];	
	TMDS_mod10 <= (TMDS_mod10==4'd8) ? 4'd0 : TMDS_mod10+4'd2;
end

ODDR2 #(.DDR_ALIGNMENT("C0"), .INIT(1'b0), .SRTYPE("ASYNC")) ODDR2_RED (.Q(TMDS_2), .D0(TMDS_shift_red[0]), .D1(TMDS_shift_red[1]), .C0(clk_TMDS), .C1(clk_TMDSn), .CE(1'b1), .R(1'b0), .S(1'b0));
ODDR2 #(.DDR_ALIGNMENT("C0"), .INIT(1'b0), .SRTYPE("ASYNC")) ODDR2_GREEN (.Q(TMDS_1), .D0(TMDS_shift_green[0]), .D1(TMDS_shift_green[1]), .C0(clk_TMDS), .C1(clk_TMDSn), .CE(1'b1), .R(1'b0), .S(1'b0));
ODDR2 #(.DDR_ALIGNMENT("C0"), .INIT(1'b0), .SRTYPE("ASYNC")) ODDR2_BLUE (.Q(TMDS_0), .D0(TMDS_shift_blue[0]), .D1(TMDS_shift_blue[1]), .C0(clk_TMDS), .C1(clk_TMDSn), .CE(1'b1), .R(1'b0), .S(1'b0));
ODDR2 #(.DDR_ALIGNMENT("C0"), .INIT(1'b0), .SRTYPE("ASYNC")) ODDR2_CLK (.Q(TMDS_clock), .D0(TMDS_shift_clock[0]), .D1(TMDS_shift_clock[1]), .C0(clk_TMDS), .C1(clk_TMDSn), .CE(1'b1), .R(1'b0), .S(1'b0));

/*
assign TMDSp[2]=clk_TMDS?TMDS_shift_red[0]:TMDS_shift_red[1];
assign TMDSn[2]=~TMDSp[2];
assign TMDSp[1]=clk_TMDS?TMDS_shift_green[0]:TMDS_shift_green[1];
assign TMDSn[1]=!TMDSp[1];
assign TMDSp[0]=clk_TMDS?TMDS_shift_blue[0]:TMDS_shift_blue[1];
assign TMDSn[0]=!TMDSp[0];
assign TMDSp_clock=(TMDS_mod10==4)?!clk_TMDS:(TMDS_mod10>5);
assign TMDSn_clock=!TMDSp_clock;
*/


endmodule
////////////////////////////////////////////////////////////////////////
// TMDS encoder
// Used to encode HDMI/DVI video data
////////////////////////////////////////////////////////////////////////

module TMDS_encoder(
	input clk,
	input pixclk_en,

	input [7:0] VD,  // video data (red, green or blue)
	input [1:0] CD,  // control data
	input VDE,  // video data enable, to choose between CD (when VDE=0) and VD (when VDE=1)
	output reg [9:0] TMDS
);

reg [3:0] balance_acc;
reg [8:0] q_m;
reg [1:0] CD2;
reg VDE2;

initial begin
	balance_acc=0;
	q_m=0;
	CD2=0;
	VDE2=0;
end

function [3:0] balance;
	input [7:0] qm;
	begin
		balance = qm[0] + qm[1] + qm[2] + qm[3] + qm[4] + qm[5] + qm[6] + qm[7] - 4'd4;
	end
endfunction

always @(posedge clk)
begin
	if(pixclk_en) begin
		// Cycle 1
		if ((VD[0] + VD[1] + VD[2] + VD[3] + VD[4] + VD[5] + VD[6] + VD[7])>(VD[0]?4'd4:4'd3)) begin
			q_m <= {1'b0,~^VD[7:0],^VD[6:0],~^VD[5:0],^VD[4:0],~^VD[3:0],^VD[2:0],~^VD[1:0],VD[0]};
		end else begin
			q_m <= {1'b1, ^VD[7:0],^VD[6:0], ^VD[5:0],^VD[4:0], ^VD[3:0],^VD[2:0], ^VD[1:0],VD[0]};
		end
		VDE2 <= VDE;
		CD2 <= CD;
		// Cycle 2
		if (VDE2) begin
			if (balance(q_m)==0 || balance_acc==0) begin
				if (q_m[8]) begin
					TMDS <= {1'b0, q_m[8], q_m[7:0]};
					balance_acc <= balance_acc+balance(q_m);
				end else begin
					TMDS <= {1'b1, q_m[8], ~q_m[7:0]};
					balance_acc <= balance_acc-balance(q_m);
				end
			end else begin
				if (balance(q_m)>>3 == balance_acc[3]) begin
					TMDS <= {1'b1, q_m[8], ~q_m[7:0]};
					balance_acc <= balance_acc+q_m[8]-balance(q_m);
				end else begin
					TMDS <= {1'b0, q_m[8], q_m[7:0]};
					balance_acc <= balance_acc-(~q_m[8])+balance(q_m);
				end
			end
		end else begin
			balance_acc <= 0;
			TMDS <= CD2[1] ? (CD2[0] ? 10'b1010101011 : 10'b0101010100) : (CD2[0] ? 10'b0010101011 : 10'b1101010100);
		end
	end
end

endmodule

////////////////////////////////////////////////////////////////////////
// TERC4 Encoder
// Used to encode the HDMI data packets such as audio
////////////////////////////////////////////////////////////////////////

module TERC4_encoder(
	input clk,
	input clk_en,
	input [3:0] data,
	output reg [9:0] TERC
);

reg [9:0] TERC_pre;

initial
begin
	TERC_pre=0;
end

always @(posedge clk)
begin

	if(clk_en) begin
		// Cycle 1
		case (data)
			4'b0000: TERC_pre <= 10'b1010011100;
			4'b0001: TERC_pre <= 10'b1001100011;
			4'b0010: TERC_pre <= 10'b1011100100;
			4'b0011: TERC_pre <= 10'b1011100010;
			4'b0100: TERC_pre <= 10'b0101110001;
			4'b0101: TERC_pre <= 10'b0100011110;
			4'b0110: TERC_pre <= 10'b0110001110;
			4'b0111: TERC_pre <= 10'b0100111100;
			4'b1000: TERC_pre <= 10'b1011001100;
			4'b1001: TERC_pre <= 10'b0100111001;
			4'b1010: TERC_pre <= 10'b0110011100;
			4'b1011: TERC_pre <= 10'b1011000110;
			4'b1100: TERC_pre <= 10'b1010001110;
			4'b1101: TERC_pre <= 10'b1001110001;
			4'b1110: TERC_pre <= 10'b0101100011;
			4'b1111: TERC_pre <= 10'b1011000011;
		endcase
		// Cycle 2
		TERC <= TERC_pre;
	end
end

//one cycle version
/*
always @(posedge clk)
begin
	if(clk_en) begin
		// Cycle 1
		case (data)
			4'b0000: TERC <= 10'b1010011100;
			4'b0001: TERC <= 10'b1001100011;
			4'b0010: TERC <= 10'b1011100100;
			4'b0011: TERC <= 10'b1011100010;
			4'b0100: TERC <= 10'b0101110001;
			4'b0101: TERC <= 10'b0100011110;
			4'b0110: TERC <= 10'b0110001110;
			4'b0111: TERC <= 10'b0100111100;
			4'b1000: TERC <= 10'b1011001100;
			4'b1001: TERC <= 10'b0100111001;
			4'b1010: TERC <= 10'b0110011100;
			4'b1011: TERC <= 10'b1011000110;
			4'b1100: TERC <= 10'b1010001110;
			4'b1101: TERC <= 10'b1001110001;
			4'b1110: TERC <= 10'b0101100011;
			4'b1111: TERC <= 10'b1011000011;
		endcase

	end
end
*/
endmodule

////////////////////////////////////////////////////////////////////////

